import filecmp
from copy import deepcopy
from difflib import SequenceMatcher
from pathlib import Path

import numpy as np
import pytest
from gradio_client import utils as client_utils

import gradio as gr
from gradio import processing_utils, utils
from gradio.data_classes import FileData
from gradio.media import get_audio


class TestAudio:
    @pytest.mark.asyncio
    async def test_component_functions(self, gradio_temp_dir, media_data):
        """
        Preprocess, postprocess serialize, get_config, deserialize
        type: filepath, numpy, file
        """
        x_wav = FileData(path=media_data.BASE64_AUDIO["path"])
        audio_input = gr.Audio()
        output1 = audio_input.preprocess(x_wav)
        assert isinstance(output1, tuple)
        assert output1[0] == 8000
        assert output1[1].shape == (8046,)

        x_wav = await processing_utils.async_move_files_to_cache([x_wav], audio_input)
        x_wav = x_wav[0]
        audio_input = gr.Audio(type="filepath")
        output1 = audio_input.preprocess(x_wav)
        assert isinstance(output1, str)
        assert Path(output1).name.endswith("audio_sample.wav")

        audio_input = gr.Audio(label="Upload Your Audio")
        assert audio_input.get_config() == {
            "autoplay": False,
            "sources": ["upload", "microphone"],
            "name": "audio",
            "buttons": ["download", "share"],
            "streaming": False,
            "show_label": True,
            "label": "Upload Your Audio",
            "container": True,
            "editable": True,
            "min_width": 160,
            "scale": None,
            "elem_id": None,
            "elem_classes": [],
            "visible": True,
            "playback_position": 0,
            "value": None,
            "interactive": None,
            "proxy_url": None,
            "type": "numpy",
            "format": None,
            "recording": False,
            "streamable": False,
            "waveform_options": {
                "sample_rate": 44100,
                "show_recording_waveform": True,
                "skip_length": 5,
                "waveform_color": None,
                "waveform_progress_color": None,
                "trim_region_color": None,
            },
            "_selectable": False,
            "key": None,
            "preserved_by_key": ["value"],
            "loop": False,
            "subtitles": None,
        }
        assert audio_input.preprocess(None) is None

        audio_input = gr.Audio(type="filepath")
        assert isinstance(audio_input.preprocess(x_wav), str)
        with pytest.raises(ValueError):
            gr.Audio(type="unknown")  # type: ignore

        rng = np.random.default_rng()
        # Confirm Audio can be instantiated with a numpy array
        gr.Audio((100, rng.random(size=(1000, 2))), label="Play your audio")

        # Output functionalities
        y_audio = client_utils.decode_base64_to_file(
            deepcopy(media_data.BASE64_AUDIO)["data"]
        )
        audio_output = gr.Audio(type="filepath")
        assert filecmp.cmp(
            y_audio.name,
            audio_output.postprocess(y_audio.name).model_dump()["path"],  # type: ignore
        )
        assert audio_output.get_config() == {
            "autoplay": False,
            "name": "audio",
            "buttons": ["download", "share"],
            "streaming": False,
            "show_label": True,
            "label": None,
            "container": True,
            "editable": True,
            "min_width": 160,
            "recording": False,
            "scale": None,
            "elem_id": None,
            "elem_classes": [],
            "visible": True,
            "playback_position": 0,
            "value": None,
            "interactive": None,
            "proxy_url": None,
            "type": "filepath",
            "format": None,
            "streamable": False,
            "sources": ["upload", "microphone"],
            "waveform_options": {
                "sample_rate": 44100,
                "show_recording_waveform": True,
                "skip_length": 5,
                "waveform_color": None,
                "waveform_progress_color": None,
                "trim_region_color": None,
            },
            "_selectable": False,
            "key": None,
            "preserved_by_key": ["value"],
            "loop": False,
            "subtitles": None,
        }

        output1 = audio_output.postprocess(y_audio.name).model_dump()  # type: ignore
        output2 = audio_output.postprocess(Path(y_audio.name)).model_dump()  # type: ignore
        assert output1 == output2

    def test_default_value_postprocess(self, media_data):
        x_wav = deepcopy(media_data.BASE64_AUDIO)
        audio = gr.Audio(value=x_wav["path"])
        assert utils.is_in_or_equal(audio.value["path"], audio.GRADIO_CACHE)

    def test_in_interface(self, media_data):
        def reverse_audio(audio):
            sr, data = audio
            return (sr, np.flipud(data))

        iface = gr.Interface(reverse_audio, "audio", "audio")
        reversed_file = iface(get_audio("audio_sample.wav"))
        reversed_reversed_file = iface(reversed_file)
        reversed_reversed_data = client_utils.encode_url_or_file_to_base64(
            reversed_reversed_file
        )
        similarity = SequenceMatcher(
            a=reversed_reversed_data, b=media_data.BASE64_AUDIO["data"]
        ).ratio()
        assert similarity > 0.99

    def test_in_interface_as_output(self):
        """
        Interface, process
        """

        def generate_noise(duration):
            return 48000, np.random.randint(-256, 256, (duration, 3)).astype(np.int16)

        iface = gr.Interface(generate_noise, "slider", "audio")
        assert iface(100).endswith(".wav")

    def test_prepost_process_to_mp3(self, gradio_temp_dir, media_data):
        x_wav = FileData(
            path=processing_utils.save_base64_to_cache(
                media_data.BASE64_MICROPHONE["data"], cache_dir=gradio_temp_dir
            )
        )
        audio_input = gr.Audio(type="filepath", format="mp3")
        output = audio_input.preprocess(x_wav)
        assert isinstance(output, str)
        assert output.endswith("mp3")
        output = audio_input.postprocess(
            (48000, np.random.randint(-256, 256, (5, 3)).astype(np.int16))
        ).model_dump()  # type: ignore
        assert output["path"].endswith("mp3")

    def test_postprocess_http_url_like(self):
        audio = gr.Audio()
        output = audio.postprocess("https://test.com/test.mp3?token=123")
        assert isinstance(output, FileData) and output.path.endswith(
            "test.mp3?token=123"
        )

    @pytest.mark.asyncio
    async def test_combine_stream_audio(self, gradio_temp_dir, media_data):
        x_wav = FileData(
            path=processing_utils.save_base64_to_cache(
                media_data.BASE64_MICROPHONE["data"], cache_dir=gradio_temp_dir
            )
        )
        bytes_output = [Path(x_wav.path).read_bytes()] * 2
        output = await gr.Audio().combine_stream(
            bytes_output, desired_output_format="wav"
        )
        assert str(output.path).endswith("wav")

        output = await gr.Audio().combine_stream(
            bytes_output, desired_output_format=None
        )
        assert str(output.path).endswith("mp3")


def test_duration_validator():
    assert gr.validators.is_audio_correct_length((8000, np.zeros((8000,))), 1, 2)[
        "is_valid"
    ]
    assert not gr.validators.is_audio_correct_length((8000, np.zeros((8000,))), 2, 3)[
        "is_valid"
    ]
    assert not gr.validators.is_audio_correct_length(
        (8000, np.zeros((8000,))), 0.25, 0.75
    )["is_valid"]



import shutil
import textwrap
from pathlib import Path

import pytest

from gradio.cli.commands.components.build import _build
from gradio.cli.commands.components.create import _create, _create_utils
from gradio.cli.commands.components.install_component import _get_executable_path
from gradio.cli.commands.components.publish import _get_version_from_file
from gradio.cli.commands.components.show import _show
from gradio.utils import core_gradio_components

core = [
    c.__name__
    for c in core_gradio_components()
    if not getattr(c, "is_template", False)
    and c.__name__ not in ["Tab", "Form", "FormComponent", "BrowserState", "Navbar"]
]


@pytest.mark.serial
@pytest.mark.flaky
@pytest.mark.parametrize(
    "template",
    core,
)
def test_template_override_component(template, tmp_path):
    """When you add a new component this test will likely fail locally
    because the js files have not been moved to the _frontend_code directory.

    Just build the python package (python -m build -w) to move the latest state of the js directory to _frontend_code.
    """
    _create(
        "MyComponent",
        tmp_path,
        template=template,
        overwrite=True,
        install=False,
        configure_metadata=False,
    )
    app = (tmp_path / "demo" / "app.py").read_text()
    component_files = _create_utils._get_component_code(template)
    answer = textwrap.dedent(
        f"""
import gradio as gr
from gradio_mycomponent import MyComponent

{component_files.demo_code.format(name="MyComponent")}

if __name__ == "__main__":
    demo.launch()
"""
    )
    assert app.strip() == answer.strip()
    assert (tmp_path / "backend" / "gradio_mycomponent" / "mycomponent.py").exists()
    source_code = (
        tmp_path / "backend" / "gradio_mycomponent" / "mycomponent.py"
    ).read_text()
    assert "@document()" not in source_code


def test_get_executable_path():
    assert _get_executable_path(
        "pip", None, "--pip-path", check_3=True
    ) == shutil.which("pip3")
    assert _get_executable_path("pip", None, "--pip-path") == shutil.which("pip")
    assert _get_executable_path(
        "pip", shutil.which("pip"), "--pip-path"
    ) == shutil.which("pip")
    assert _get_executable_path(
        "gradio", None, "--pip-path", check_3=True
    ) == shutil.which("gradio")
    with pytest.raises(
        ValueError,
        match=r"Could not find foo. Please ensure it is installed and in your PATH or pass the --foo-path parameter.",
    ):
        _get_executable_path("foo", None, "--foo-path")
    with pytest.raises(
        ValueError,
        match=r"The provided foo path \(/foo/bar/fum\) does not exist or is not a file.",
    ):
        _get_executable_path("foo", "/foo/bar/fum", "--foo-path")


def test_raise_error_component_template_does_not_exist(tmp_path):
    with pytest.raises(
        ValueError,
        match="Cannot find NonExistentComponent in gradio.components, gradio.layouts, or gradio._simple_templates",
    ):
        _create(
            "MyComponent",
            tmp_path,
            template="NonExistentComponent",
            overwrite=True,
            install=False,
            configure_metadata=False,
        )


def test_do_not_replace_class_name_in_import_statement(tmp_path):
    _create(
        "MyImage",
        template="Image",
        directory=tmp_path,
        overwrite=True,
        install=False,
        configure_metadata=False,
    )
    code = (tmp_path / "backend" / "gradio_myimage" / "myimage.py").read_text()
    assert "import PIL.Image" in code
    assert "class MyImage" in code
    assert "PIL.Image.Image" in code


def test_raises_if_directory_exists(tmp_path):
    with pytest.raises(
        Exception
    ):  # Keeping it a general exception since the specific exception seems to differ between operating systems
        _create("MyComponent", tmp_path, configure_metadata=False)


def test_show(capsys):
    _show()
    stdout, _ = capsys.readouterr()
    assert "Form Component" in stdout
    assert "Beginner Friendly" in stdout
    assert "Layout" in stdout
    assert "Dataframe" not in stdout
    assert "Dataset" not in stdout


@pytest.mark.xfail
@pytest.mark.parametrize("template", ["Image"])
def test_build(template, virtualenv):
    # Copy pnpm-lock.yaml to not cause unintended changes tracked by git
    pnpm_lock = Path(__file__).parent / ".." / "pnpm-lock.yaml"
    pnpm_copy = Path(__file__).parent / ".." / "pnpm-lock-copy.yaml"
    shutil.copy(str(pnpm_lock), str(pnpm_copy))

    # Using the js/preview/test directory will use the workspace code
    dir_ = (
        Path(__file__).parent / ".." / "js" / "preview" / "test" / "testtextbox"
    ).resolve()
    shutil.rmtree(str(dir_), ignore_errors=True)

    try:
        # Local installs of gradio and gradio-client
        gradio_dir = Path(__file__).parent / ".."
        client = Path(__file__).parent / ".." / "client" / "python"
        virtualenv.run("pip install build")
        virtualenv.run(f"pip install -e {str(gradio_dir)}")
        virtualenv.run(f"pip install -e {str(client)}")

        virtualenv.run(
            f"{shutil.which('gradio')} cc create TestTextbox --template {template} --no-configure-metadata --directory {str(dir_)}",
        )
        assert (dir_ / "frontend" / "node_modules").exists()

        # need to reinstall local client because installing the custom component
        # will pull latest stable version from pypi
        virtualenv.run(f"pip install -e {str(client)}")
        virtualenv.run(f"{shutil.which('gradio')} cc build {str(dir_)}")

        template_dir: Path = dir_ / "backend" / "gradio_testtextbox" / "templates"
        assert template_dir.exists() and template_dir.is_dir()
        assert list(template_dir.glob("**/index.js"))
        assert (dir_ / "dist").exists() and list((dir_ / "dist").glob("*.whl"))
    finally:
        shutil.move(str(pnpm_copy), str(pnpm_lock))
        shutil.rmtree(str(dir_), ignore_errors=True)


def test_build_fails_if_component_not_installed(tmp_path):
    _create(
        "MyComponent",
        tmp_path,
        template="SimpleTextbox",
        overwrite=True,
        install=False,
        configure_metadata=False,
    )
    with pytest.raises(
        ValueError,
        match=r"Your custom component package \(gradio_mycomponent\) is not installed!",
    ):
        _build(tmp_path)


def test_fallback_template_app(tmp_path):
    _create(
        "SimpleComponent2",
        directory=tmp_path,
        overwrite=True,
        install=False,
        configure_metadata=False,
    )
    app = (tmp_path / "demo" / "app.py").read_text()
    answer = textwrap.dedent(
        """

import gradio as gr
from gradio_simplecomponent2 import SimpleComponent2


with gr.Blocks() as demo:
    gr.Markdown("# Change the value (keep it JSON) and the front-end will update automatically.")
    SimpleComponent2(value={"message": "Hello from Gradio!"}, label="Static")


if __name__ == "__main__":
    demo.launch()

"""
    )
    assert app.strip() == answer.strip()


def test_get_version_from_wheel():
    assert (
        _get_version_from_file(Path("gradio_textwithattachments-0.0.3-py3-none.whl"))
        == "0.0.3"
    )
    assert (
        _get_version_from_file(Path("gradio_textwithattachments-1.0.3b12-py3-none.whl"))
        == "1.0.3b12"
    )
